/* This file is part of the Amalthea library.
 *
 * Copyright (C) 2018-2021 Eugene 'Vindex' Stulin
 *
 * Distributed under the Boost Software License 1.0 or (at your option)
 * the GNU Lesser General Public License 3.0 or later.
 */

module amalthea.csv;

public import amalthea.libcore;
import amalthea.dataprocessing : getIndex;

import std.algorithm, std.array, std.path, std.string;


/*******************************************************************************
 * Structure for working with CSV-tables.
 */
struct CSV {
    private string csvFilePath;
    private char delimiter = ',';

    private string[][] dataTable;
    string[] headersOfColumns;

    /***************************************************************************
     * The constructor takes a path to CSV-file.
     */
    this(string filepath) {
        csvFilePath = filepath;
        generateDataTable();
    }

    /***************************************************************************
     * The constructor takes a two-dimensional array as the initiating table.
     */
    this(string[][] table) {
        createTable(table);
    }

    /***************************************************************************
     * The constructor takes a path to future CSV-file and
     * two-dimensional array as the initiating table.
     * Params:
     *     filepath = Path to file for saving CSV table. The existence of
     *                the file, as well as its contents, do not matter.
     *     table    = 2D array with content for the CSV file.
     */
    this(string filepath, string[][] table) {
        csvFilePath = filepath;
        createTable(table);
    }

    /***************************************************************************
     * The constructor takes an array of table column names
     * and a two-dimensional array as the initiating table.
     * Params:
     *     keys = Table column titles.
     *     table = 2D  with content for the CSV file (without titles).
     */
    this(string[] keys, string[][] table) {
        createTable(keys, table);
    }

    /***************************************************************************
     * The method returns content of the table of the CSV-object
     * as two-dimensional array.
     */
    string[][] getTable() {
        return dataTable.dup;
    }

    /***************************************************************************
     * The method returns rows of the table as associative arrays,
     * where keys are names of the columns and values are content of the fields.
     */
    string[string][] getArrayOfRowMaps() {
        string[][] dataTable = getTable[1 .. $];
        string[string][] arrayOfMaps;
        foreach(row; dataTable) {
            string[string] map;
            foreach(i, header; headersOfColumns) {
                map[header] = row[i];
            }
            arrayOfMaps ~= map;
        }
        return arrayOfMaps;
    }

    /***************************************************************************
     * Find the table field by another field in the same row (key1 and value)
     * and column name (key2).
     */
    string getValueOfField(string key1, string value, string key2) {
        string result;
        auto arrayOfMaps = getArrayOfRowMaps();
        foreach(mapLine; arrayOfMaps) {
            if (mapLine[key1] != value) continue;
            result = mapLine[key2];
            break;
        }
        return result;
    }

    /***************************************************************************
     * The method finds table row by field (by column name and value),
     * returns associative array, where keys are names of columns and
     * values are content of fields of this row.
     */
    string[string] getRowMapByField(string key1, string value) {
        string[string] result;
        auto arrayOfMaps = getArrayOfRowMaps();
        foreach(mapLine; arrayOfMaps) {
            if (mapLine[key1] != value) continue;
            result = mapLine;
            break;
        }
        return result;
    }

    /***************************************************************************
     * The method sets new value (value2)
     * by other field in the same row and column name.
     */
    void setValueOfField(string key1, string value1,
                         string key2, string value2) {
        ssize_t index = -1;
        auto arrayOfMaps = getArrayOfRowMaps();
        foreach(i, mapLine; arrayOfMaps) {
            if (mapLine[key1] != value1) continue;
            index = i;
            break;
        }
        if (index == -1) return;
        dataTable[index+1][headersOfColumns.getIndex(key2)] = value2;
    }

    /***************************************************************************
     * The methods adds new table row.
     */
    void addRow(string[] row) {
        if (headersOfColumns.length > 0)
            row.length = headersOfColumns.length;
        dataTable ~= row;
    }
    void addRow(S...)(S args) {
        string[] row;
        foreach(arg; args) row ~= to!string(arg);
        addRow(row);
    }

    /***************************************************************************
     * The method deletes table row by field (by its key and value).
     */
    void deleteRowByKeyAndValue(string key, string value) {
        ssize_t index = -1;
        auto arrayOfMaps = getArrayOfRowMaps();
        foreach(i, mapLine; arrayOfMaps) {
            if (mapLine[key] != value) continue;
            index = i;
            break;
        }
        if (index == -1) return;
        ++index;
        if (index != dataTable.length-1) {
            dataTable = dataTable[0 .. index] ~ dataTable[index+1 .. $];
        } else {
            dataTable = dataTable[0 .. index];
        }
    }
    alias deleteRowByField = deleteRowByKeyAndValue;

    /***************************************************************************
     * The method writes table content to the initiating file.
     */
    void rewrite() {
        writeTo(csvFilePath);
    }

    /***************************************************************************
     * The method writes table content to the specified file.
     */
    void writeTo(string filepath) {
        auto dir = std.path.dirName(filepath);
        if (dir != "" && !dir.exists) mkdirRecurse(dir);
        auto f = File(filepath, "w");
        string newContent;
        foreach(row; dataTable) {
            foreach(i, field; row) {
                newContent ~= `"`~field.replace(`"`, `""`)~`"`;
                if (i != row.length-1) newContent ~= delimiter;
            }
            newContent ~= "\n";
        }
        f.write(newContent);
        f.flush();
    }

    /***************************************************************************
     * The two-dimensional array will be used as the content of the CSV table.
     */
    void createTable(string[][] table) {
        dataTable = table.dup;
    }

    /***************************************************************************
     * The two-dimensional array will be used as the content of the CSV table.
     * Params:
     *     keys = Titles for columns.
     *     table = 2D array for CSV fields.
     */
    void createTable(string[] keys, string[][] table) {
        dataTable = table.dup;
        dataTable = keys ~ table;
        headersOfColumns = keys.dup;
    }

    private void generateDataTable() {
        if (!csvFilePath.exists) return;
        auto f = File(csvFilePath, "r");
        string[][] arr;
        size_t i;
        foreach(line; f.byLine) {
            arr.length++;
            foreach(el; splitCSVLine(line.idup)) {
                arr[i] ~= [el];
            }
            ++i;
        }
        if (i == 0) return;
        dataTable = arr.dup;
        string[] keys;
        foreach(headerField; dataTable[0]) {
            keys ~= headerField;
        }
        headersOfColumns = keys.dup;
    }

    private string[] splitCSVLine(string csvRow) {
        string[] tmp = csvRow.split(delimiter);
        string[] result;
        for(size_t i; i < tmp.length; i++) {
            auto el = tmp[i];
            if (el.startsWith('"')) {
                el = el[1 .. $];
                if (el.endsWith('"')) {
                    el = el[0 .. $-1];
                } else {
                    auto nextIndex = i+1;
                    foreach(fragment; tmp[nextIndex .. $]) {
                        i++;
                        el ~= delimiter ~ fragment;
                        if (el.endsWith('"')) {
                            el = el[0 .. $-1];
                            break;
                        }
                    }
                }
            }
            result ~= el;
        }
        foreach(ref el; result) {
            el = el.replace(`""`, `"`);
        }
        return result;
    }

}
